import asyncio
from threading import Thread
import websockets
import cv2
import numpy as np
import base64
import json
from sil_sdk.client.vis_async_client import VISAsyncClient

class VISModule:
    def __init__(self, server_uri="ws://localhost:4201", start_server=True):
        self.server_uri = server_uri
        self.loop = asyncio.new_event_loop()
        self.results = {}
        self.running_tasks = {}
        self.background_thread = Thread(target=self._start_loop, daemon=True)
        self.background_thread.start()
        if start_server:
            self._run_async_task(self._start_websocket_server()).result()
        self.client = VISAsyncClient(server_uri)

    def _start_loop(self):
        asyncio.set_event_loop(self.loop)
        self.loop.run_forever()

    def _run_async_task(self, coro):
        return asyncio.run_coroutine_threadsafe(coro, self.loop)

    async def _start_websocket_server(self):
        async def handler(websocket, path):
            try:
                async for message in websocket:
                    request = json.loads(message)
                    cmd = request.get("command")

                    if cmd == "obj_detection_load":
                        await asyncio.sleep(1)
                        await websocket.send(json.dumps({"status": "loaded", "module": "obj_detection"}))

                    elif cmd == "obj_detection_run":
                        await asyncio.sleep(1)
                        dummy_img = np.zeros((100, 100, 3), dtype=np.uint8)
                        _, buffer = cv2.imencode('.jpg', dummy_img)
                        img_base64 = base64.b64encode(buffer).decode("utf-8")
                        await websocket.send(json.dumps({
                            "status": "done",
                            "module": "obj_detection",
                            "image": img_base64,
                            "objects": [{"label": "dummy", "box": [10, 10, 50, 50]}]
                        }))

                    elif cmd == "obj_detection_stop":
                        await websocket.send(json.dumps({"status": "stopped", "module": "obj_detection"}))

            except websockets.ConnectionClosed:
                pass

        host, port = self._parse_uri(self.server_uri)
        self.ws_server = await websockets.serve(handler, host, port)

    def _parse_uri(self, uri):
        uri = uri.replace("ws://", "")
        if "/" in uri:
            uri = uri.split("/", 1)[0]
        host, port = uri.split(":")
        return host, int(port)

    def load(self, modules):
        for module in modules:
            self._run_async_task(self.client.load_module(module)).result()

    def run(self, module, prompt=None, bbox=None, mask=None):
        if module in self.running_tasks:
            return
        task = self._run_async_task(self._run_and_store_result(module, prompt, bbox, mask))
        self.running_tasks[module] = task

    async def _run_and_store_result(self, module, prompt, bbox, mask):
        result = await self.client.run_module(module, prompt, bbox, mask)
        self.results[module] = result

    def get_result(self, module):
        return self.results.get(module)

    def stop_all(self):
        for module in list(self.running_tasks.keys()):
            self._run_async_task(self.client.stop_module(module)).result()
        self.running_tasks.clear()

    def close(self):
        self.stop_all()
        self.loop.call_soon_threadsafe(self.loop.stop)
        self._run_async_task(self.client.close()).result()



import asyncio
import websockets
import json
import threading


class DAGModule:
    def __init__(self, server_uri="ws://localhost:8767"):
        self.server_uri = server_uri
        self.websocket = None
        self.status = {}
        self.loop = asyncio.new_event_loop()
        self.thread = threading.Thread(target=self._run_loop, daemon=True)
        self.thread.start()

    def _run_loop(self):
        asyncio.set_event_loop(self.loop)
        self.loop.run_until_complete(self._connect())

    async def _connect(self):
        self.websocket = await websockets.connect(self.server_uri)
        asyncio.create_task(self._listen())  # Background listener

    async def _listen(self):
        while True:
            try:
                response = await self.websocket.recv()
                self.status = json.loads(response)
            except websockets.ConnectionClosed:
                print("[DAG] WebSocket disconnected. Reconnecting...")
                await self._connect()
                break

    def grasp(self, grasp_mode="pinch", velocity=75,
              pos_min=None, pos_max=None, max_contact_force=None):
        # Default joint configs if not provided
        default_pos = {f"joint_{i}": 10 for i in range(1, 13)}
        default_max = {f"joint_{i}": 80 for i in range(1, 13)}
        default_force = {f"joint_{i}": 50 for i in range(1, 13)}

        message = {
            "grasp_command": True,
            "enable_multi_grasp": False,
            "grasp_mode": grasp_mode,
            "pos_min": pos_min or default_pos,
            "pos_max": pos_max or default_max,
            "velocity": velocity,
            "max_contact_force": max_contact_force or default_force
        }
        asyncio.run_coroutine_threadsafe(self.websocket.send(json.dumps(message)), self.loop)

    def get_current_status(self):
        return self.status

    def close(self):
        if self.websocket:
            asyncio.run_coroutine_threadsafe(self.websocket.close(), self.loop)



import asyncio
import websockets
import json
import threading


class SARModule:
    def __init__(self, server_uri="ws://localhost:8766"):
        self.server_uri = server_uri
        self.websocket = None
        self.status = {}
        self.loop = asyncio.new_event_loop()
        self.thread = threading.Thread(target=self._run_loop, daemon=True)
        self.thread.start()

    def _run_loop(self):
        asyncio.set_event_loop(self.loop)
        self.loop.run_until_complete(self._connect())

    async def _connect(self):
        self.websocket = await websockets.connect(self.server_uri)
        asyncio.create_task(self._listen())  # Start background feedback listener

    async def _listen(self):
        while True:
            try:
                response = await self.websocket.recv()
                self.status = json.loads(response)
            except websockets.ConnectionClosed:
                print("[SAR] WebSocket disconnected. Reconnecting...")
                await self._connect()
                break

    def set_parameters(self, graspCommand=False, velocity=50, strokeMin=0, strokeMax=100,
                       maxContactForce=50, axialThreshold=50, lateralThreshold=50):
        message = {
            "graspCommand": graspCommand,
            "setVelocity": velocity,
            "strokeMin": strokeMin,
            "strokeMax": strokeMax,
            "setMaxContactForce": maxContactForce,
            "setAxialForceThreshold": axialThreshold,
            "setLateralForceThreshold": lateralThreshold
        }
        asyncio.run_coroutine_threadsafe(self.websocket.send(json.dumps(message)), self.loop)

    def get_current_status(self):
        return self.status

    def close(self):
        if self.websocket:
            asyncio.run_coroutine_threadsafe(self.websocket.close(), self.loop)



import asyncio
import json
import websockets
import cv2
import base64
import numpy as np

class VISAsyncClient:
    def __init__(self, server_uri="ws://localhost:4201"):
        self.server_uri = server_uri
        self.websocket = None
        self.results = {}
        self.response_queue = asyncio.Queue()

    async def connect(self):
        if self.websocket is None:
            self.websocket = await websockets.connect(self.server_uri)
            asyncio.create_task(self.listen_for_results())

    async def listen_for_results(self):
        while True:
            try:
                response = await self.websocket.recv()
                data = json.loads(response)

                if "image" in data:
                    img_base64 = data["image"]
                    try:
                        img_data = base64.b64decode(img_base64)
                        np_arr = np.frombuffer(img_data, np.uint8)
                        img = cv2.imdecode(np_arr, cv2.IMREAD_COLOR)
                        data["image"] = img
                    except Exception as e:
                        print("Error decoding image:", e)
                        data["image"] = None

                module_name = data.get("module", "unknown")
                self.results[module_name] = data
                await self.response_queue.put(data)
            except websockets.ConnectionClosed:
                print("Connection lost. Reconnecting...")
                self.websocket = None
                await asyncio.sleep(1)
                await self.connect()

    async def load_module(self, module_name):
        await self.connect()
        request = {"command": f"{module_name}_load"}
        await self.websocket.send(json.dumps(request))

    async def run_module(self, module_name, prompt=None, bbox=None, mask=None):
        await self.connect()
        request = {"command": f"{module_name}_run"}
        if prompt:
            request["prompt"] = prompt
        if bbox:
            request["bbox"] = bbox
        if mask is not None:
            _, buffer = cv2.imencode('.png', mask)
            request["mask"] = base64.b64encode(buffer).decode('utf-8')
        await self.websocket.send(json.dumps(request))

        while True:
            data = await self.response_queue.get()
            if data.get("module") == module_name:
                return data

    async def stop_module(self, module_name):
        await self.connect()
        request = {"command": f"{module_name}_stop"}
        await self.websocket.send(json.dumps(request))

    async def close(self):
        if self.websocket:
            await self.websocket.close()



# dag_async_client.py

import asyncio
import json
import websockets

class DAGAsyncClient:
    def __init__(self, server_uri="ws://<server-ip>:8767"):
        self.server_uri = server_uri
        self.websocket = None
        self.response_queue = asyncio.Queue()

    async def connect(self):
        if self.websocket is None:
            self.websocket = await websockets.connect(self.server_uri)
            asyncio.create_task(self.listen_for_responses())

    async def listen_for_responses(self):
        while True:
            try:
                response = await self.websocket.recv()
                data = json.loads(response)
                await self.response_queue.put(data)
            except websockets.ConnectionClosed:
                print("DAG connection lost. Reconnecting...")
                self.websocket = None
                await self.connect()
                break

    async def send_command(self, command):
        await self.connect()
        await self.websocket.send(json.dumps(command))

    async def get_response(self):
        return await self.response_queue.get()

    async def close(self):
        if self.websocket:
            await self.websocket.close()
